"""Contains SECS I message class."""
from __future__ import annotations

import secsgem.common

from .header import SecsIHeader


class SecsIBlock(secsgem.common.Block[SecsIHeader]):
    """Data block for SECS I."""

    header_type = SecsIHeader
    length_format = "B"
    checksum_format = "H"


class SecsIMessage(secsgem.common.Message[SecsIBlock]):
    """Class for SECS I message.

    Contains all required data and functions.
    """

    block_size = 244
    block_type = SecsIBlock

    @property
    def header(self) -> SecsIHeader:
        """Get the header."""
        return self._blocks[-1].header

    @property
    def data(self) -> bytes:
        """Get the data."""
        return b"".join(block.data for block in self._blocks)

    @property
    def complete(self) -> bool:
        """Check if the message is complete."""
        return self.blocks[-1].header.last_block



"""Class for stream 05 function 10."""

from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F10(SecsStreamFunction):
    """exception post - confirm.

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F10
        Header only

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F10()
        S5F10 .

    """

    _stream = 5
    _function = 10

    _data_format = None

    _to_host = False
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""SECS-I settings class."""
from __future__ import annotations

import secsgem.common


class SecsISettings(secsgem.common.Settings):
    """Settings for SECS-I connection.

    These attributes can be initialized in the constructor and accessed as property.

    Example:
        >>> import secsgem.secsi
        >>>
        >>> settings = secsgem.secsi.SecsISettings(port="COM1")
        >>> settings.port
        'COM1'
        >>> settings.speed
        9600

    """

    def __init__(self, **kwargs) -> None:
        """Initialize settings."""
        super().__init__(**kwargs)

        self._port = kwargs.get("port", "")
        self._speed = kwargs.get("speed", 9600)

    @property
    def port(self) -> str:
        """Serial port.

        Default: ""
        """
        return self._port

    @property
    def speed(self) -> int:
        """Serial port baud rate.

        Default: 9600
        """
        return self._speed

    def create_protocol(self) -> secsgem.common.Protocol:
        """Protocol class for this configuration."""
        from .protocol import SecsIProtocol  # pylint: disable=import-outside-toplevel

        return SecsIProtocol(self)

    def create_connection(self) -> secsgem.common.Connection:
        """Connection class for this configuration."""
        return secsgem.common.SerialConnection(self)

    @property
    def name(self) -> str:
        """Name of this configuration."""
        return f"SECSI-{self.port}"

    def generate_thread_name(self, functionality: str) -> str:
        """Generate a unique thread name for this configuration and a provided functionality.

        Args:
            functionality: name of the functionality to generate thread name for

        Returns:
            generated thread name

        """
        return f"secsgem_SecsI_{functionality}_{self.port}@{self.speed}"



"""Header for the secs messagess."""
from __future__ import annotations

import struct
import typing

import secsgem.common


class SecsIHeader(secsgem.common.Header):
    """Generic SECS I header.

    Base for different specific headers
    """

    length = 10

    def __init__(  # pylint: disable=too-many-arguments
            self,
            system: int,
            session_id: int,
            stream: int = 0,
            function: int = 0,
            block: int = 0,
            from_equipment: bool = False,
            require_response: bool = False,
            last_block: bool = True):
        """Initialize a SECS I header.

        Args:
            system: message ID
            session_id: device / session ID
            stream: stream
            function: function
            block: block number
            from_equipment: message is send from equipment
            require_response: response requested
            last_block: last block of multi block message

        Example:
            >>> import secsgem.secsi
            >>>
            >>> secsgem.secsi.SecsIHeader(3, 100)
            SecsIHeader({session_id:0x0064, stream:00, function:00, system:0x00000003, block:0x0000, from_host:False, \
require_response:False, last_block:True})

        """
        super().__init__(system, session_id, stream, function, require_response)

        self._block = block
        self._from_equipment = from_equipment
        self._last_block = last_block

    @property
    def block(self) -> int:
        """Get block number."""
        return self._block

    @property
    def from_equipment(self) -> bool:
        """Get if the message is from equipment (True) or host (False).

        This is the reverse bit (r-bit).
        """
        return self._from_equipment

    @property
    def last_block(self) -> bool:
        """Get if the message is the last of a multi-block message.

        This is the end bit (e-bit).
        """
        return self._last_block

    @property
    def _as_dictionary(self) -> dict[str, typing.Any]:
        """Get the data as dictionary.

        Returns:
            Header data as dictionary.

        """
        return {
            "system": self._system,
            "session_id": self._session_id,
            "stream": self._stream,
            "function": self._function,
            "block": self._block,
            "from_equipment": self._from_equipment,
            "require_response": self._require_response,
            "last_block": self._last_block,
        }

    def __str__(self) -> str:
        """Generate string representation for an object of this class."""
        return (
            "{"
            f"session_id:0x{self.session_id:04x}, "
            f"stream:{self.stream:02d}, "
            f"function:{self.function:02d}, "
            f"system:0x{self.system:08x}, "
            f"block:0x{self.block:04x}, "
            f"from_host:{self.from_equipment!r}, "
            f"require_response:{self.require_response!r}, "
            f"last_block:{self.last_block!r}"
            "}"
        )

    def __repr__(self) -> str:
        """Generate textual representation for an object of this class."""
        return f"{self.__class__.__name__}({self.__str__()})"

    def encode(self) -> bytes:
        """Encode header to SECS I message.

        Returns:
            encoded header

        Example:
            >>> import secsgem.secsi
            >>> import secsgem.common
            >>>
            >>> header = secsgem.secsi.SecsIHeader(2, 100)
            >>> secsgem.common.format_hex(header.encode())
            '00:64:00:00:80:00:00:00:00:02'

        """
        session_id = self.session_id
        if self.from_equipment:
            session_id |= 0b1000000000000000

        stream = self.stream
        if self.require_response:
            stream |= 0b10000000

        block = self.block
        if self.last_block:
            block |= 0b1000000000000000

        return struct.pack(
            ">HBBHI",
            session_id,
            stream,
            self.function,
            block,
            self.system,
        )

    @classmethod
    def decode(cls, data: bytes) -> SecsIHeader:
        """Decode data to SecsIHeader object.

        Args:
            data: data to decode

        Returns:
            new header object

        """
        res = struct.unpack(">HBBHI", data)

        return SecsIHeader(
            res[4],
            res[0] & 0b0111111111111111,
            res[1] & 0b01111111,
            res[2],
            res[3] & 0b0111111111111111,
            (((res[0] & 0b1000000000000000) >> 15) == 1),
            (((res[1] & 0b10000000) >> 7) == 1),
            (((res[3] & 0b1000000000000000) >> 15) == 1),
        )



"""SECS-I protocol implementation."""
from __future__ import annotations

import typing

import secsgem.common

from .header import SecsIHeader
from .message import SecsIBlock, SecsIMessage

if typing.TYPE_CHECKING:
    from ..secs.functions.base import SecsStreamFunction
    from .settings import SecsISettings


class SecsIProtocol(secsgem.common.Protocol[SecsIMessage, SecsIBlock]):
    """Implementation for SECS-I protocol."""

    ENQ = 0b00000101
    EOT = 0b00000100
    ACK = 0b00000110
    NAK = 0b00010101

    block_size = 244

    message_type = SecsIMessage

    def __init__(self, settings: SecsISettings):
        """Instantiate SECS I protocol class.

        Args:
            settings: protocol and communication settings

        Example:
            import secsgem.secsi

            settings = secsgem.secsi.SecsISettings(
                port="COM1",
            )

            def onConnect(event, data):
                print ("Connected")

            client = secsgem.secsi.SecsIProtocol(settings)
            client.events.connected += onConnect

            client.enable()

            time.sleep(3)

            client.disable()

        """
        super().__init__(settings)
        self._settings: SecsISettings = settings

    def _create_message_for_function(
            self,
            function: SecsStreamFunction,
            system_id: int,
    ) -> secsgem.common.Message:
        """Create a protocol specific message for a function.

        Args:
            function: function to create message for
            system_id: system

        Returns:
            created message

        """
        return SecsIMessage(
            SecsIHeader(
                system_id,
                self._settings.session_id,
                function.stream,
                function.function,
                require_response=function.is_reply_required,
                from_equipment=(self._settings.device_type == secsgem.common.DeviceType.EQUIPMENT),
            ),
            function.encode(),
        )

    def serialize_data(self) -> dict[str, typing.Any]:
        """Return data for serialization.

        :returns: data to serialize for this object
        :rtype: dict
        """
        return {"port": self._settings.port, "baud_rate": self._settings.speed}

    def _on_connected(self, _: dict[str, typing.Any]):
        """Handle connection was established event."""
        self._thread.start()
        self.events.fire("connected", {"connection": self})
        self.events.fire("communicating", {"connection": self})

    def _on_disconnected(self, _: dict[str, typing.Any]):
        """Handle connection was _ event."""
        # clear receive buffer
        self.events.fire("disconnected", {"connection": self})

        self._thread.stop()

        self._receive_buffer.clear()

    def _on_disconnecting(self, _: dict[str, typing.Any]):
        pass

    def _process_send_queue(self):
        if self._send_queue.empty():
            return

        while not self._send_queue.empty():
            self._connection.send_data(bytes([self.ENQ]))

            enq_resonse = self._receive_buffer.wait_for_byte(peek=True)

            if enq_resonse == self.ENQ and self._settings.device_type == secsgem.common.DeviceType.HOST:
                self._process_received_data()
                continue

            enq_resonse = self._receive_buffer.pop_byte()

            block_info = self._send_queue.get()

            self._connection.send_data(block_info.data)

            data_response = self._receive_buffer.wait_for_byte()

            block_info.resolve(data_response == self.ACK)

    def _process_received_data(self):
        if len(self._receive_buffer) < 1:
            return

        while len(self._receive_buffer) > 0:
            receive_byte = self._receive_buffer.pop_byte()

            if receive_byte != self.ENQ:
                self._logger.info("Expected ENQ, received '%s'. Ignoring", receive_byte)

            self._connection.send_data(bytes([self.EOT]))

            length = self._receive_buffer.wait_for_byte(peek=True)

            data = self._receive_buffer.wait_for(length + 3)

            response = SecsIBlock.decode(data)

            if response is None:
                self._connection.send_data(bytes([self.NAK]))
                return

            # redirect message to hsms handler
            self._thread.queue_block(self, response)

            self._connection.send_data(bytes([self.ACK]))

    def _on_connection_message_received(self, source: object, message: SecsIMessage):
        """Message received from connection.

        Args:
            source: source of event
            message: received data message

        """
        decoded_message = self._settings.streams_functions.decode(message)
        self._communication_logger.info("< %s\n%s", message, decoded_message, extra=self._get_log_extra())

        # someone is waiting for this message
        if message.header.system in self._response_queues:
            self._response_queues[message.header.system].put_nowait(message)
        else:
            self.events.fire("message_received", {"connection": source, "message": message})

    def _get_log_extra(self) -> dict[str, typing.Any]:
        """Get extra fields for logging."""
        return {"port": self._settings.port,
                "speed": self._settings.speed,
                "session_id": self._settings.session_id,
                "remoteName": self._settings.name}



"""Contains objects and functions to create and handle serial connections."""
from __future__ import annotations

import logging
import threading
import time
import typing

import serial

from .connection import Connection
from .helpers import format_hex

if typing.TYPE_CHECKING:
    from .settings import Settings


class SerialConnection(Connection):  # pylint: disable=too-many-instance-attributes
    """Connection class used for serial connections."""

    _receiver_timeout = 0.5

    def __init__(self, settings: Settings):
        """Initialize a serial connection.

        Args:
            settings: protocol and communication settings

        """
        super().__init__(settings)

        self._logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)
        self._bytestream_logger = logging.getLogger("bytestream")

        self.__port: serial.Serial | None = None

        self._enabled = False
        self._receiver_thread_running = False
        self._stop_receiver_thread = False
        self._receiver_thread: threading.Thread | None = None

    @property
    def _port(self) -> serial.Serial:
        if self.__port is None:
            raise ConnectionError(f"SecsI port is not connected: {self}")

        return self.__port

    def __str__(self):
        """Get the contents of this object as a string."""
        return (
            f"Serial connection to "
            f"{self._settings.port}@{self._settings.speed}"
            f" session_id={self._settings.session_id}"
        )

    def enable(self):
        """Enable the connection.

        Open port and start receiver thread.
        """
        # only start if not already enabled
        if self._enabled:
            return

        # mark connection as enabled
        self._enabled = True

        self.__port = serial.Serial(self._settings.port, self._settings.speed, timeout=self._receiver_timeout)

        # start data receiving thread
        self._receiver_thread = threading.Thread(
            target=self._receiver_thread_function,
            args=(),
            name=f"secsgem_secsIConnection_receiver_{self._settings.port}@{self._settings.speed}",
        )
        self._receiver_thread.daemon = True
        self._receiver_thread.start()

        try:
            self.on_connected({"source": self})
        except Exception:  # pylint: disable=broad-except
            self._logger.exception("ignoring exception for on_connected handler")

        # wait until thread is running
        while not self._receiver_thread_running:
            pass

    def disable(self):
        """Disable the connection.

        Close port and stop receiver thread.
        """
        # only stop if enabled
        if not self._enabled:
            return

        # mark connection as disabled
        self._enabled = False

        self._stop_receiver_thread = True

        # wait for connection thread to stop
        while self._stop_receiver_thread:
            time.sleep(0.2)

    def _receiver_thread_function(self):
        """Thread for receiving incoming data and sending it to the protocol handler."""
        self._receiver_thread_running = True

        try:
            self._receiver_loop()
        except Exception:  # pylint: disable=broad-except
            self._logger.exception("exception")

        # notify listeners of disconnection
        try:
            self.on_disconnecting({"source": self})
        except Exception:  # pylint: disable=broad-except
            self._logger.exception("ignoring exception for on_disconnecting handler")

        # close the socket
        self._port.close()

        # notify listeners of disconnection
        try:
            self.on_disconnected({"source": self})
        except Exception:  # pylint: disable=broad-except
            self._logger.exception("ignoring exception for on_disconnected handler")

        # reset all flags
        self._connected = False
        self._receiver_thread_running = False
        self._stop_receiver_thread = False

    def _receiver_loop(self):
        # check if shutdown requested
        while not self._stop_receiver_thread:
            data = self._port.read(self._port.in_waiting) if self._port.in_waiting > 0 else self._port.read()

            if len(data) > 0:
                self._bytestream_logger.debug("< %s", format_hex(data))
                self.on_data({"source": self, "data": data})

    def send_data(self, data: bytes) -> bool:
        """Send data to the remote host.

        Args:
            data: encoded data.

        Returns:
            True if succeeded, False if failed

        """
        self._bytestream_logger.debug("> %s", format_hex(data))
        self._port.write(data)

        return True



"""module imports."""

from .header import SecsIHeader
from .protocol import SecsIProtocol
from .settings import SecsISettings

__all__ = [
    "SecsIHeader",
    "SecsIProtocol",
    "SecsISettings",
]
